-- Copyright (c) 2012 Takahiro Shinozaki
-- 
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.

--------------------------------------------------------------------------------
-- |
-- Module      :  WFST
-- Copyright   :  Takahiro Shinozaki 2012
-- License     :  MIT
-- 
-- Maintainer  :  www.ailab.tj.chiba-u.jp/~shinot
-- Stability   :  experimental
-- Portability :  portable
-- 
-- A weighted finite state transducer (WFST) module. 
--------------------------------------------------------------------------------

{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE BangPatterns #-}

-- WFST Assumptions --------------------------------------
-- 1) the initial state is the start state of the first arc in the file
-- 2) there is only one final state 
-- 3) Final state does not have associated cost
-- 4) arcs are sorted by its start node ID.

module WFST (Arc, -- A structure representing an arc of WFST with a set of interface functions
             WFST, -- Structure representing a WFST with a set of interface functions
             Path,
             PathList,
             initArc, -- Makes and initializes a new arc
             arcSt, -- Start state of an arc
             arcEd, -- End note of an arc
             inLbl,  -- Input label
             outLbl,  -- Output label
             arcWght,  -- Weight
             finSt,
             iniSt,
             nSt,
             inspectWFST,
             wfstFromList,
             expandPathListWithILblScore,
             expandPathListForEps,
             pathListToList,
             pathListFromList,
             pathFromList,
             pathToList,
             pathEd,
             pathScore,
             pathListSize,
             pathListTake,
             pathListTakeWhileScore,
             wfstFromFile,
             readWFST,
             outSymFromFile
             ) 
       where

import GHC.Exts
import Data.Maybe
import qualified Data.Vector as V
import qualified Data.Map as M
import qualified Data.Set as S
import Data.Int (Int64)
-- import qualified Data.List.Stream as List
import qualified Data.List as List
import MHUtils
import qualified Data.ByteString.Lazy.Char8 as L8

import Control.Monad.ST
import qualified Data.HashTable.ST.Cuckoo as H

-- Arc structure definition
data Arc = Arc {
  arcSt :: !Int,
  arcEd :: !Int,
  inLbl :: !Int,
  outLbl :: !Int,
  arcWght :: !Double,
  arcAuxS :: !Double, -- A general purpose auxiliary score variable
  arcAuxN :: !Int -- A scaffold to avoid exchanging complex data structures between C and haskell
  } deriving (Eq)
           
data Path = Path {
     pathArcs :: [Arc], -- ordered from end to start
     pathSt :: !Int,
     pathEd :: !Int,
     pathScore :: !Double
     }

type PathList = [Path]

-- WFST structure definition
data WFST = WFSTElem {
  numSt :: !Int,
  initalSt :: !Int,
  finalSt :: !Int, 
  startIdx :: V.Vector Int,
  dataBody :: V.Vector Arc -- dataBody[0] is a dummy item for states having no outbound arc
  }
          | WFSTComposit { -- Not implemented yet
  numSt :: Int,
  initalSt :: Int,
  finalSt :: Int, 
  operation :: Int,
  elems :: [WFST]
  }

wfstStateArcs :: WFST -> Int -> V.Vector Arc
wfstStateArcs wfst i 
  = V.slice six len (dataBody wfst)
  where six = (startIdx wfst) V.! i
        len = arcAuxN $ (dataBody wfst) V.! six

initArc s e i o w = Arc (fromIntegral s) (fromIntegral e) (fromIntegral i) (fromIntegral o) (realToFrac w) 0.0 0
nullArc = Arc 0 0 0 0 0.0 0.0 0
setArcAuxN (Arc s e i o w as an) newan = Arc s e i o w as newan
setArcAuxS (Arc s e i o w as an) newas = Arc s e i o w newas an

arcIsEpsilon a = (inLbl a == 0)

-- extends a path by one arc at the end, with some additional weight
pathExtend :: Path -> Arc -> Double -> Path
pathExtend (Path as pst _ sc) a auxW =
  Path (a:as) pst (arcEd a) (sc + arcWght a + auxW)

finSt w = finalSt w
iniSt w = initalSt w
nSt w = numSt w

instance Show Arc where
  show (Arc s e i o w auxn auxs) 
--    = "Arc " ++ (show s) ++ " " ++ (show e) ++ " " ++ (show i) ++ " " ++ (show o) ++ " " ++ (show w)
    = "Arc " ++ (show s) ++ " " ++ (show e) ++ " " ++ (show i) ++ " " ++ (show o) ++ " " ++ (show w) ++ " " ++ (show auxn) ++ " " ++ (show auxs)
instance Show WFST where
  show (WFSTElem n ini fin stidx dbdy) 
    = "NumSt=" ++ (show n) ++ " " ++ "InitSt=" ++ (show ini) ++ " " ++ "FinalSt=" ++ (show fin) ++ "\n" ++
--       (List.foldr (\x y -> x++"\n"++y) "" $ tail (List.map show (V.toList dbdy))) ++
       (List.foldr (\x y -> x++"\n"++y) "" (List.map show (V.toList dbdy))) ++
       show stidx
  

-- Makes a wfst from a list. The start state of the first arc is treated as the initial state.
-- The final state is a start state of an arc that has negative end state. There must be one 
-- and only one end state. The arcs must be sorted by start state id with increasing order.
-- Non-used state id may exist but the largest state id value determins internal 
-- start array size.
wfstFromList :: [Arc] -> WFST 
wfstFromList [] = WFSTElem 0 0 0 V.empty V.empty
wfstFromList aclist = WFSTElem numst initSt finSt stidx arcArray
  where arcArray = V.fromList $
                   emptyDummyArc : setStartGroupNum (-1, 0, [], [], emptyDummyArc) aclist -- The first item is a dummy
        emptyDummyArc = (Arc (-1) (-1) (-1) (-1) 0.0 0.0 0)
        -- Number of arcs that has the same start state id. The number is given when the start id changes in the list
        -- Ex. If arc start is like [2,2,3,3,3,6,6,9], then [2,0,3,0,0,2,0,1].
        setStartGroupNum :: (Int,Int,[Int],[Arc], Arc) -> [Arc] -> [Arc]
        setStartGroupNum (pre, !num, ntmp, atmp, fin) [] = List.zipWith setArcAuxN (List.reverse atmp) (num:ntmp) List.++ [fin]
        setStartGroupNum (pre, !num, ntmp, atmp, fin) (x:xs) 
          | (arcEd x) < 0 = setStartGroupNum (pre, num, ntmp, atmp, x) xs
          | pre == (-1) = setStartGroupNum ((arcSt x), 1, [], [x], fin) xs
          | pre == (arcSt x) = setStartGroupNum (pre, num+1, 0:ntmp, x:atmp, fin) xs
          | otherwise = block List.++ setStartGroupNum ((arcSt x), 1, [], [x], fin) xs 
            where block = List.zipWith setArcAuxN (List.reverse atmp) (num:ntmp) 
        arcstmp = V.unsafeTail arcArray
        numst = 1 + (fromIntegral $ arcSt $ V.maximumBy larger arcstmp)
          where larger a b | diff < 0 = LT
                           | diff > 0 = GT
                           | diff == 0 = EQ
                  where diff = (arcSt a) - (arcSt b)
        initSt = (fromIntegral $ arcSt $ V.head arcstmp)::Int
        finSt = arcSt (arcstmp V.! (V.length arcstmp -1))
        -- Arc start position index. (startIndex[s] is the start position of arcs in wfst that has start id s)
        -- States that do not have outbound arc is mapped to the dummy arc arcArray[0].
        -- Ex. If arc start is like [0,1,2,2,3,3,6,6,9], then [1,2,3,5,0,0,7,0,0,9].
        stidx = V.fromList $ startIndex (-1, 1) (V.take ((V.length arcstmp) -1) arcstmp)
        startIndex :: (Int, Int) -> V.Vector Arc -> [Int]
        startIndex (pre, pos) v | V.length v == 0  = if pre < (fromIntegral finSt) 
                                                      then (replicate (finSt-(fromIntegral pre)) 0) 
                                                      else []
                                | pre == (arcSt x) = startIndex (pre, pos+1) xs
                                | (arcSt x)-pre == 1 = pos : startIndex ((arcSt x), pos+1) xs
                                | otherwise = (replicate (fromIntegral ((arcSt x)-pre-1)) 0) 
                                              List.++ (pos:(startIndex ((arcSt x), pos+1) xs))
          where x = V.unsafeHead v
                xs = V.unsafeTail v



inspectWFST :: WFST -> Double
inspectWFST (WFSTElem n ini fin stidx arcs) = 
  V.foldl' (\acc ac -> acc + (arcWght ac) + (realToFrac (arcSt ac)) + (realToFrac (arcEd ac))) 0.0 arcs

-- Expands a set of WFST states with initial scores
-- Input labels are mapped to scores
expandPathListWithILblScore :: WFST -> PathList -> (Int -> Double) -> PathList
expandPathListWithILblScore wfst pl ilbl2score 
  = concatMap doSrcPath pl
  where
    doSrcPath :: Path -> PathList
    doSrcPath p =
      mapMaybe doArc $ V.toList arcs
      where arcs = wfstStateArcs wfst (pathEd p)
        
            doArc a | arcIsEpsilon a = Nothing
            doArc a = Just $ pathExtend p a (ilbl2score $ inLbl a)

-- Expands a PathList for epsilon transitions
expandPathListForEps :: WFST -> PathList -> PathList

-- Functional Map-based version
-- expandPathListForEps wfst pl 
--   = M.elems $ foldl doPath M.empty pl
--   where
--     doPath :: M.Map Int Path -> Path -> M.Map Int Path
--     doPath m p = 
--       case updateIfSmaller m p of
--         Nothing -> m
--         Just m2 ->
--           let arcs = wfstStateArcs wfst (pathEd p) in
--           V.foldl (doArc p) m2 arcs

--     doArc :: Path -> M.Map Int Path -> Arc -> M.Map Int Path
--     doArc p m a | not $ arcIsEpsilon a  = m -- not epsilon
--     doArc p m a | arcIsEpsilon a =
--       doPath m $ pathExtend p a 0.0

--     updateIfSmaller :: M.Map Int Path -> Path -> Maybe (M.Map Int Path)
--     updateIfSmaller m p =
--       let e = pathEd p
--           sc = pathScore p
--       in
--         case M.lookup e m of
--           Nothing -> Just $ M.insert e p m
--           Just op ->
--             if sc < pathScore op
--             then Just $ M.insert e p m
--             else Nothing 


-- A version with ST and HashTables
expandPathListForEps wfst pl
  = runST $ do
       ht <- H.newSized (2 * length pl)
       mapM_ (doPath ht) pl
       hToValueList ht

  where
     doPath :: H.HashTable s Int Path -> Path -> ST s ()
     doPath ht p = do
       nv <- updateIfSmaller ht (pathEd p) p
       if nv then 
          let arcs = wfstStateArcs wfst (pathEd p) in
          mapM_ (doArc ht p) (V.toList arcs)
       else
          return ()

     doArc :: H.HashTable s Int Path -> Path -> Arc -> ST s ()
     doArc ht p a | not $ arcIsEpsilon a = return ()
     doArc ht p a | arcIsEpsilon a = 
       doPath ht $ pathExtend p a 0.0

     updateIfSmaller :: H.HashTable s Int Path -> Int -> Path -> ST s Bool
     updateIfSmaller ht ix p = do
       mov <- H.lookup ht ix
       case mov of
         Nothing -> do
           H.insert ht ix p
           return True
         Just ov ->
           if (pathScore p) < (pathScore ov)
           then do
             H.insert ht ix p
             return True
           else return False

     hToValueList :: H.HashTable s k a -> ST s [a]
     hToValueList ht = H.foldM (\pl (key, nv) -> return $ nv:pl) [] ht

pathListToList :: PathList -> [Path]
pathListToList = id
  
pathListFromList :: [Path] -> PathList                                
pathListFromList = id

pathToList :: Path -> [Arc]
pathToList p = List.reverse $ pathArcs p

pathFromList :: [Arc] -> Path
pathFromList [] = Path [] 0 0 0
pathFromList as =
  let ras = reverse as in
  Path ras (arcSt $ head as) (arcEd $ head ras) (foldl (+) 0 $ map arcWght as)

pathListSize :: PathList -> Int
pathListSize = length
  
pathListTake :: Int -> PathList -> PathList
pathListTake = take

pathListTakeWhileScore :: (Double -> Bool) -> PathList -> PathList
pathListTakeWhileScore pred pl = 
  takeWhile (pred . pathScore) pl
  
wfstFromFile :: String -> IO WFST
wfstFromFile fname = do {buf <- L8.readFile fname; return (readWFST buf)}
  
readWFST :: L8.ByteString -> WFST
readWFST buf = wfstFromList arclist
  where arclist = List.map (line2arc . L8.words) lines
        lines = List.filter (\x -> x /= L8.empty) $ L8.splitWith (=='\n') buf
        line2arc [s,e,i,o,w] = initArc (mBtoI s) (mBtoI e) (mBtoI i) (mBtoI o) (mBtoD w)
        line2arc [s,e,i,o]   = initArc (mBtoI s) (mBtoI e) (mBtoI i) (mBtoI o) 0.0
        line2arc [s]         = initArc (mBtoI s) (-1) 0 0 0.0


-- Output ID to symbol mapping definitions
outSymFromFile :: String -> IO (V.Vector L8.ByteString)
outSymFromFile fname = do {buf <- L8.readFile fname; return (readOutSym buf)}

readOutSym :: L8.ByteString -> V.Vector L8.ByteString
readOutSym buf = V.fromList $ L8.splitWith (=='\n') buf

